from .RoPE import RoPE
import torch.nn as nn
from dataclasses import dataclass
from typing import Optional
import torch
from .activations import Dynamic_erf, DynamicTanh
from torch.nn.attention.flex_attention import flex_attention
from functools import partial


@dataclass
class BertAttentionArgs:
    n_heads:int = 16
    n_kv_heads:int = 4
    dim:int = 128*16
    soft_cap:Optional[int] = 20
    device:str = 'cuda:0'
    layer_norm:str = "RMSNorm"

class BertAttention(nn.Module):
    def __init__(self, args: BertAttentionArgs, rope:RoPE, score_mod:callable=None):
        super().__init__()

        self.args = args

        self.n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
        self.n_heads_q = args.n_heads
        self.n_rep = self.n_heads_q // self.n_kv_heads
        self.head_dim = args.dim // args.n_heads

        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")

        self.wo = nn.Linear(args.n_heads * self.head_dim, args.dim, bias=False)

        self.wq = nn.Linear(args.dim, args.n_heads * self.head_dim, bias=False)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * self.head_dim, bias=False)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * self.head_dim, bias=False)
        
        self.hq_dim = self.n_heads_q * self.head_dim
        self.hkv_dim = self.n_kv_heads * self.head_dim

        self.rope = rope

        self.attn = partial(flex_attention, score_mod=score_mod, enable_gqa= self.n_heads_q != self.n_kv_heads)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len, dim = x.shape

        x = self.layer_norm(x)

        xq = self.wq(x)
        xk = self.wk(x)
        xv = self.wv(x)

        xq = xq.view(batch_size, seq_len, self.n_heads_q, self.head_dim)
        xk = xk.view(batch_size, seq_len, self.n_kv_heads, self.head_dim)
        xv = xv.view(batch_size, seq_len, self.n_kv_heads, self.head_dim)

        xq, xk = self.rope.apply_rotary_embeddings(xq, xk)

        # (b, seq, h_q, head_dim) -> (b, h_q, seq, head_dim)
        xq = xq.transpose(1,2)
        xk = xk.transpose(1,2)
        xv = xv.transpose(1,2)

        out = self.attn(xq, xk, xv)
        
        out = out.transpose(1,2).contiguous().view(batch_size, seq_len, dim) # (b, seq, dim)
        return self.wo(out) #(b, seq, dim)



import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from .dataloader import JSONLDataset
from functools import partial
from dataclasses import dataclass, field
from torch.optim import Optimizer, AdamW, SGD
from pathlib import Path
from .hf_hub import push_model_to_hub
from huggingface_hub import hf_hub_download
import os
from safetensors.torch import load_file
from .modules import get_args_from_hub, cache_or_fetch
import time
import threading
from progress_table import ProgressTable
import numpy as np
from .model import Cirilla
from ..LLM_pieces import get_activation
try:
    from megablocks.layers.router import clear_router_zloss
    from megablocks.layers.moe import clear_load_balancing_loss
    use_megablocks = True
except ImportError:
    use_megablocks = False
    pass
import re

@dataclass
class TrainingArgs:
    n_epoch:int = 100
    optim:Optimizer = AdamW
    use_muon_optim:bool = False
    static_triton_graph:bool = False
    lr:float = 5e-5
    batch_size:int = 4
    valid_every_n:int=5
    save_local_async:bool = False
    init_method_str:str = 'xavier_uniform_'
    local_checkpoint_folder:Path = './test_model'
    fuse_optim:bool = True
    optim_kwargs:dict[str,str] = field(default_factory=lambda: {'fused':True, 'foreach':False})

    renew_training:bool = True
    save_checkpoint_n_iterations:int = None
    save_checkpoint_min:int = 2

    push_checkpoint_to_hub:bool = False
    push_checkpoint_to_hub_n_local_saves:int = 4
    
    hf_repo_id:str = None
    private_hf_repo:bool=True
    hf_tags:list[str] = field(default_factory=lambda: ["pytorch", "text-generation", "moe", "custom_code"])
    hf_license:str = 'mit'
    languages:list[str] = field(default_factory=lambda: ["en"])
    model_card:str = None

    @property
    def stateful_optim(self):
        if self.optim == SGD:
            return False
        return True
    
    @property
    def init_method(self):
        return getattr(torch.nn.init, self.init_method_str)

class CirillaTrainer:
    def __init__(self, model:nn.Module, training_args:TrainingArgs):
        self.model = model
        self.args = training_args
        self.optims_to_save = None
        self.pulled_from_hub = False
        self.optim = self._prepare_optimizer(**training_args.optim_kwargs)
        if self.args.use_muon_optim:
            self.moun_optim = get_activation("motif-technologies/optimizer")
            
        self.criterion = nn.CrossEntropyLoss(ignore_index=1, # tokenizer.convert_tokens_to_ids(padding token) ; by default it's 1
                                            label_smoothing=0.1)

        self.n_checkpoints = 0

        print(f'n trainable params: {(model.n_params/1e6):.2f} M')

    def train(self, dataset:JSONLDataset, valid_dataset:JSONLDataset=None):

        dataset_len = cache_or_fetch('DATA_LEN', dataset.path_signature)

        if dataset_len % self.args.batch_size == 0 and self.args.static_triton_graph:
            print('Using static triton graph')
            static_training = True
        else:
            static_training = False
        
        dataset_path = dataset.path_signature

        skip_n_data_points = cache_or_fetch('N_DATA_POINTS', dataset_path)
        if skip_n_data_points is None:
            skip_n_data_points = 0

        dataloader = DataLoader(dataset, shuffle=False, batch_size=self.args.batch_size)
        n_iter_total = self.args.n_epoch * len(dataset) - skip_n_data_points
        del dataset

        if valid_dataset is not None:
            valid_dataloader = DataLoader(valid_dataset, shuffle=False, batch_size=self.args.batch_size)
            valid_dataset = 1

        start_time = time.time()

        self._set_global_vars()

        n_iter = -1
        loaded_checkpoint = False

        os.makedirs(self.args.local_checkpoint_folder, exist_ok=True)

        state_type = "stateful" if self.args.stateful_optim else "non-stateful"
        optimizer_path = os.path.join(self.args.local_checkpoint_folder, "optimizer_states.pt")
        model_path = os.path.join(self.args.local_checkpoint_folder, "model.pt")

        if self.args.renew_training and not self.pulled_from_hub:
            if os.path.exists(optimizer_path) or not self.args.stateful_optim:
                if os.path.exists(model_path):
                    self._load_local_checkpoint()
                    loaded_checkpoint = True
                else:
                    if skip_n_data_points > 0:
                        raise FileNotFoundError(
                            f"Couldn't find model path at: {model_path}"
                        )
            else:
                if skip_n_data_points > 0:
                    raise FileNotFoundError(
                        f"Couldn't find optimizer states path for a {state_type} optimizer at: {optimizer_path}"
                    )

        if not loaded_checkpoint and not self.pulled_from_hub:
            print("Training from scratch")

            if self.args.init_method_str is not None:
                self._weights_init()

            if self.args.fuse_optim:
                self._fuse_optim()

        if static_training:
            self._set_prior_training_vars()

        prev_mean_loss = 10

        def loss_color(loss:list):
            nonlocal prev_mean_loss
            mean_loss = np.mean(loss)

            if mean_loss < prev_mean_loss * 0.95:
                color = "green"
            elif mean_loss < prev_mean_loss * 1.05:
                color = "yellow"
            else:
                color = "red"
            
            prev_mean_loss = mean_loss
            return color
        
        prev_mean_loss_valid = 10

        def loss_color_valid(loss:list):
            if len(loss) > 0:
                nonlocal prev_mean_loss_valid
                mean_loss = np.mean(loss)

                if mean_loss < prev_mean_loss_valid * 0.95:
                    color = "green"
                elif mean_loss < prev_mean_loss_valid * 1.05:
                    color = "yellow"
                else:
                    color = "red"
                
                prev_mean_loss_valid = mean_loss
                return color

            
        prev_mean_time = 1
                        
        def time_color(times:list):
            nonlocal prev_mean_time
            mean_time = np.mean(np.diff(times))

            if mean_time < prev_mean_time * 1.1:
                color =  "green"
            elif mean_time < prev_mean_time * 1.5:
                color = "yellow"
            else:
                color = "red"

            prev_mean_time = mean_time
            return color
            
        ptable = ProgressTable(
                pbar_show_progress=False,
                pbar_show_throughput=False,
                pbar_show_eta=True,
                default_column_width=8,
                default_header_color="bold",
                                )
        
        main_pbar = ptable.pbar(
                        n_iter_total,
                        position=1,
                        show_progress=True,
                        style="rich alt lightmagenta_ex lightwhite_ex",
                    )
        
        for epoch in range(self.args.n_epoch):

            if skip_n_data_points // dataset_len > 0:
                skip_n_data_points -= dataset_len
                n_iter += dataset_len // self.args.batch_size
                continue

            times = [time.time()]
            losses = []
            v_losses = []
            
            ptable['epoch'] = epoch

            self.model.train()

            for data in dataloader:

                n_iter += 1

                if n_iter * self.args.batch_size < skip_n_data_points:
                    continue

                loss_item = self.training_step(data)

                losses.append(loss_item)
                times.append(time.time())

                ptable.update('train loss', round(loss_item, 3), aggregate='mean', color='cyan')
                ptable.update('time', round(times[-1] - times[-2], 3), aggregate='mean', color='blue')

                do_checkpoint, push_hub = self._check_if_do_checkpoint(time.time() - start_time, n_iter)
                
                if do_checkpoint:
                    start_time = time.time()
                    try:
                        if self.args.save_local_async:
                            self._save_local_checkpoint_async()
                        else:
                            self._save_local_checkpoint()
                    except Exception as e:
                        sync_ = 'asynchronously' if self.args.save_local_async else 'synchronously'
                        print(f"Failed to save local checkpoint {sync_}:{e}\nSaving synchronously")
                        self._save_local_checkpoint()

                    cache_or_fetch('N_DATA_POINTS', dataset_path, (n_iter + 1) * self.args.batch_size)
                    if push_hub and self.args.push_checkpoint_to_hub:
                        local_name = dataset_path
                        try:
                            local_name = os.path.basename(local_name)
                        except:
                            local_name = local_name.replace('.jsonl', '').replace('.', '-')
                        match_obj = re.match(r'^(?:[\w-]+\/)?[\w.-]+$', local_name)
                        push_dataset_name = match_obj.group(0) if match_obj else None
                        try:
                            self._push_all_to_hub_async(loss_item, push_dataset_name)
                            
                        except Exception as e:
                            print(f"Failed to push asynchronously to HF hub: {e}\nPushing synchronously")
                            self._push_all_to_hub(loss_item, push_dataset_name)
                
                main_pbar.update(self.args.batch_size)
                
            if valid_dataset is not None:
                if epoch % self.args.valid_every_n == 0:

                    if n_iter * self.args.batch_size < skip_n_data_points:
                        continue

                    self.model.eval()

                    for data in valid_dataloader:
                        loss_item = self.inference_step(data)

                        v_losses.append(loss_item)

                        ptable.update('valid loss', round(loss_item, 3), aggregate='mean', color='lightcyan_ex')
                    
                    torch.cuda.empty_cache()
                    ptable.next_row(split=True, color={'time': time_color(times), 'train loss': loss_color(losses), 'valid loss': loss_color_valid(v_losses)})

            ptable.next_row(split=valid_dataset is None, color={'time': time_color(times), 'train loss': loss_color(losses), 'valid loss': loss_color_valid(v_losses)})

        self._save_local_checkpoint()
        if self.args.push_checkpoint_to_hub:
            self._push_all_to_hub(loss_item, dataset_path)

        cache_or_fetch('N_DATA_POINTS', dataset_path, (n_iter+1) * self.args.batch_size)

    def training_step(self, data):

        torch.compiler.cudagraph_mark_step_begin()

        out = self.model.pred(data[0])
        loss = self.criterion(out.view(-1, self.model.args.vocab_size), data[1].view(-1))

        # clear losses that will cause a memory leak
        if use_megablocks:
            clear_load_balancing_loss()
            clear_router_zloss()
        loss_item = loss.item()
        loss.backward()

        return loss_item
    
    @torch.inference_mode()
    def inference_step(self, data):
        if use_megablocks:
            clear_load_balancing_loss()
            clear_router_zloss()
        out = self.model.pred(data[0])
        loss = self.criterion(out.view(-1, self.model.args.vocab_size), data[1].view(-1))
        return loss.item()

    def benchmark(self):
        
        self._set_global_vars()

        x = torch.randint(0, self.model.args.vocab_size,
                        (4, self.model.args.context_window),
                        dtype=torch.long, device=self.model.args.device)
        
        y = torch.randint(0, self.model.args.vocab_size,
                        (4, self.model.args.context_window),
                        dtype=torch.long, device=self.model.args.device)
        
        def loss_color(distance):
            if distance < 8:
                return "green"
            elif distance < 9:
                return "yellow"
            else:
                return "red"
            
        def time_color(distance):
            if distance < 0.76:
                return "green"
            elif distance < 0.9:
                return "yellow"
            else:
                return "red"

        if self.args.xavier_init:
            self._xavier_init()

        self._fuse_optim()

        self._set_prior_training_vars()

        for i in range(5): #warm up for benchmark
            loss_item = self.training_step((x, x))

        torch.cuda.synchronize()

        ptable = ProgressTable(
                pbar_show_progress=False,
                pbar_show_throughput=False,
                pbar_show_eta=True,
                default_column_width=8,
                default_header_color="bold",
                                )
        
        main_pbar = ptable.pbar(
                        100,
                        position=1,
                        show_progress=True,
                        style="rich alt lightmagenta_ex lightwhite_ex",
                    )
        
        times = [time.time()]
        losses = []

        for i in range(100):

            if i % 5 == 0:
                ptable['epoch'] = i

            loss_item = self.training_step((x, y))

            times.append(time.time())
            losses.append(loss_item)
            
            ptable.update('train loss', loss_item, aggregate='mean', color='cyan')
            ptable.update('time', round(times[-1] - times[-2], 4), aggregate='mean', color='blue')

            if i % 5 == 0: # new row every 5 iterations
                ptable.next_row(split=True, color={'time': time_color(np.mean(np.diff(times[-5:]))), 'train loss': loss_color(np.mean(losses[-5:]))})

            main_pbar.update(1)

        ptable.close()
        print(f'average time for epoch: {np.mean(np.diff(times)):.4f}')

    def _check_if_do_checkpoint(self, time, iter_step):
        if self.args.save_checkpoint_min is not None:
            if time >= self.args.save_checkpoint_min * 60:
                self.n_checkpoints += 1
                return True, self.n_checkpoints % self.args.push_checkpoint_to_hub_n_local_saves == 0
            
        if self.args.save_checkpoint_n_iterations is not None:
            if iter_step % self.args.save_checkpoint_n_iterations == 0:
                self.n_checkpoints += 1
                return True, self.n_checkpoints % self.args.push_checkpoint_to_hub_n_local_saves == 0
            
        return False, False

    def _weights_init(self):
        for param in self.model.parameters():
            if param.dim() > 1:
                self.args.init_method(param)

    @staticmethod
    def _set_global_vars():
        torch.set_float32_matmul_precision('high')
        torch.backends.cudnn.benchmark = True
        torch._dynamo.config.capture_scalar_outputs = True
    
    @staticmethod
    def _set_prior_training_vars():
        torch._inductor.config.triton.cudagraph_skip_dynamic_graphs = True
        torch._inductor.config.triton.cudagraph_dynamic_shape_warn_limit = None

    def _prepare_optimizer(self, **optim_kwargs):
        return partial(self.args.optim, **optim_kwargs, lr=self.args.lr)
    
    def _fuse_optim(self):
        print('fusing optimizers...')

        if not self.args.use_muon_optim:

            self.optimizer_by_name = {}
            for name, p in self.model.named_parameters():
                self.optimizer_by_name[name] = self.optim([p])

            params_by_name = dict(self.model.named_parameters())
            optimizer_dict = {params_by_name[name]: opt for name, opt in self.optimizer_by_name.items()}

        else:
            get_default_muon_param_groups = self.moun_optim.muon.get_default_muon_param_groups
            muon_param_groups = get_default_muon_param_groups(self.model)

            self.optimizer_by_name = {}
            for name, p in self.model.named_parameters():

                if name in muon_param_groups[0]['names']: # matrices
                    group = {
                        "params": [p],
                        "names": [name],
                        "use_muon": True
                    }
                    self.optimizer_by_name[name] = self.moun_optim.Muon([group], lr=self.args.lr)

                else: # biases, LayerNorm weights
                    self.optimizer_by_name[name] = self.optim([p])

            params_by_name = dict(self.model.named_parameters())
            optimizer_dict = {params_by_name[name]: opt for name, opt in self.optimizer_by_name.items()}

        self._register_hooks(optimizer_dict)

    def _register_hooks(self, optimizer_dict):
        
        def optimizer_hook(parameter):
            optimizer_dict[parameter].step()
            optimizer_dict[parameter].zero_grad(set_to_none=True)

        for p in self.model.parameters():
            if p in optimizer_dict:
                p.register_post_accumulate_grad_hook(optimizer_hook)
            else:
                print(f"Unknown param of shape: {p.shape}")

    def _save_local_checkpoint(self):
        if not hasattr(self, 'optimizer_by_name') and self.args.fuse_optim:
            self._fuse_optim()
            
        torch.save(self.model.state_dict(), os.path.join(self.args.local_checkpoint_folder, 'model.pt'))

        if self.args.stateful_optim and self.args.fuse_optim:

            optim_states = {name: opt.state_dict() for name, opt in self.optimizer_by_name.items()}
            torch.save(optim_states, os.path.join(self.args.local_checkpoint_folder, 'optimizer_states.pt'))
        
        elif self.args.stateful_optim and self.optims_to_save is not None:

            optim_states = {name: opt.state_dict() for name, opt in self.optims_to_save.items()}
            torch.save(optim_states, os.path.join(self.args.local_checkpoint_folder, 'optimizer_states.pt'))

    def _save_local_checkpoint_async(self):
        def worker():
            self._save_local_checkpoint()

        threading.Thread(target=worker, daemon=True).start()

    def _load_local_checkpoint(self):
        self.model.load_state_dict(torch.load(\
            os.path.join(self.args.local_checkpoint_folder,'model.pt')))
        
        if self.args.stateful_optim and self.args.fuse_optim:

            loaded_states = torch.load(\
                os.path.join(self.args.local_checkpoint_folder, 'optimizer_states.pt'),
                map_location=self.model.args.device)

            self._load_optim_from_checkpoint(loaded_states)

        elif self.args.stateful_optim and self.optims_to_save is not None:

            loaded_states = torch.load(\
                os.path.join(self.args.local_checkpoint_folder, 'optimizer_states.pt'),
                map_location=self.model.args.device)

            for optim_name, optim in self.optims_to_save.items():
                optim.load_state_dict(loaded_states[optim_name])

        else:
            self._fuse_optim()
        
    def _load_optim_from_checkpoint(self, loaded_states):

        if self.optims_to_save is not None:
            
            for optim_name, optim in self.optims_to_save.items():
                optim.load_state_dict(loaded_states[optim_name])

            return

        params_by_name = dict(self.model.named_parameters())
        self.optimizer_by_name = {}

        if self.args.use_muon_optim:
            get_default_muon_param_groups = self.moun_optim.muon.get_default_muon_param_groups
            muon_param_groups = get_default_muon_param_groups(self.model)[0]['names']
        else:
            muon_param_groups = []
        
        for name, state in loaded_states.items():
            if name not in params_by_name:
                print(f"Skipping unknown param: {name}")
                continue

            p = params_by_name[name]

            if name in muon_param_groups:

                group = {
                    "params": [p],
                    "names": [name],
                    "use_muon": True
                }
                opt = self.moun_optim.Muon([group], lr=self.args.lr)
                opt.load_state_dict(state)
                self.optimizer_by_name[name] = opt

            else:

                opt = self.optim([p])
                opt.load_state_dict(state)
                self.optimizer_by_name[name] = opt

        optimizer_dict = {params_by_name[n]: o for n, o in self.optimizer_by_name.items()}

        self._register_hooks(optimizer_dict)

    def _push_all_to_hub(self, loss, dataset_name):
        push_model_to_hub(
            repo_id = self.args.hf_repo_id,
            model = self.model,
            loss = loss,
            dataset_name = dataset_name,
            private = self.args.private_hf_repo,
            optmizer_states_path = os.path.join(self.args.local_checkpoint_folder, 'optimizer_states.pt'),
            tags = self.args.hf_tags,
            license = self.args.hf_license,
            languages = self.args.languages,
            model_card = self.args.model_card
        )

    def _push_all_to_hub_async(self, loss, dataset_name):
        args = (loss, dataset_name)

        def worker(loss_value, dataset_name):
            self._push_all_to_hub(loss_value, dataset_name)

        t = threading.Thread(target=worker, args=args, daemon=True)
        t.start()
    
    def _pull_optim_from_hub(self):
        file_path = hf_hub_download(
            repo_id=self.args.hf_repo_id,
            filename="optimizer_states.pt",
        )

        if not os.path.exists(file_path):
            print('no optimizer states file found')

        with open(file_path, "rb") as f:
            loaded_states = torch.load(f, map_location=self.model.args.device)

        self._load_optim_from_checkpoint(loaded_states)

    def _pull_model_from_hub(self):
        model_args = self.model.args
        pulled_args = get_args_from_hub(self.args.hf_repo_id, type(self.model.args))

        if model_args != pulled_args:
            print(f"Current model args don't correspond to the HF model's args.\nCurrent args:\n{model_args}\nThe model will use the HF args:\n{pulled_args}")
            self.model = Cirilla(pulled_args)

        file_path = hf_hub_download(
            repo_id=self.args.hf_repo_id,
            filename="model.safetensors",
        )

        loaded = load_file(file_path)
        if "output.weight" not in loaded and "output.weight" in self.model.state_dict():
            loaded['output.weight'] = loaded["emb.embeddings.weight"]

        self.model.load_state_dict(loaded)

    def _pull_all_from_hub(self):
        self._pull_model_from_hub()
        self._pull_optim_from_hub()
        self.pulled_from_hub = True
        print(f'pulled from hub: {self.args.hf_repo_id}')



import os
import json

def instructions_into_conv(input_path, out_path):
    for file in os.listdir(input_path):
        with open(os.path.join(input_path, file), 'r') as f:
            qa = json.load(f)
            for k,v in qa.items():
                out = {'subject': file.split('.')[0],
                    'text': [{'role':'user', 'content': k}, {'role': 'assistant', 'content': v}],
                    'data type': 'conv',
                    'source': 'fandom'}
                
                with open(out_path, 'a') as f:
                    f.write(json.dumps(out) + '\n')



from datasets import load_dataset
import json
import re

pretraining = './training_datasets/pretraining.jsonl'
pretraining_ = './training_datasets/pretraining_overview.jsonl'

""""""
dataset = load_dataset("roneneldan/TinyStories", split='train[:2%]')

data = []

with open(pretraining_, 'a') as f:
    f.write(json.dumps({'metadata': {'dataset':"roneneldan/TinyStories", 'split':'train[:2%]'},
        'num_rows':dataset.num_rows}) + '\n')

for text in dataset:
    data.append({'text': text['text'], 'data type': 'plain text', 'metadata': {'dataset':"roneneldan/TinyStories", 'split':'train[:2%]'}})

with open(pretraining, 'a') as f:
    for d in data:
        f.write(json.dumps(d) + '\n')

""""""
dataset = load_dataset("roneneldan/TinyStoriesInstruct", split='train[:1%]')

data = []

with open(pretraining_, 'a') as f:
    f.write(json.dumps({'metadata': {'dataset':"roneneldan/TinyStoriesInstruct", 'split':'train[:1%]'},
        'num_rows':dataset.num_rows}) + '\n')
    
for text in dataset:
    text = text['text']
    if not bool(re.match(r'^(\w+:|<\|)', text)) and len(text) > 10: # if text doesn't start with some_word: or <|somw_word
        data.append({'text': text, 'data type': 'plain text', 'metadata': {'dataset':"roneneldan/TinyStoriesInstruct", 'split':'train[:1%]'}})

with open(pretraining, 'a') as f:
    for d in data:
        f.write(json.dumps(d) + '\n')

""""""

dataset = load_dataset("nyu-mll/glue", "mnli", split="train[:3%]")

with open(pretraining_, 'a') as f:
    f.write(json.dumps({'metadata': {'dataset':"nyu-mll/glue", 'subset':'mnli', 'split':'train[:3%]'},
        'num_rows':dataset.num_rows * 2}) + '\n')

data = []

for text in dataset:
    data.append({'text': text['premise'], 'data type': 'plain text', 'metadata': {'dataset':"nyu-mll/glue", "subset":"mnli", 'split':'train[:3%]'}})
    data.append({'text': text['hypothesis'], 'data type': 'plain text', 'metadata': {'dataset':"nyu-mll/glue", "subset":"mnli", 'split':'train[:3%]'}})

with open(pretraining, 'a') as f:
    for d in data:
        f.write(json.dumps(d) + '\n')

""""""

dataset = load_dataset("SimpleStories/SimpleStories", split="train[:2%]")

data = []

with open(pretraining_, 'a') as f:
    f.write(json.dumps({'metadata': {'dataset':"SimpleStories/SimpleStories", 'split':'train[:2%]'},
        'num_rows':dataset.num_rows}) + '\n')

for text in dataset:
    data.append({'text': text['story'], 'data type': 'plain text', 'metadata': {'dataset':"SimpleStories/SimpleStories", 'split':'train[:2%]'}})

with open(pretraining, 'a') as f:
    for d in data:
        f.write(json.dumps(d) + '\n')

""""""
# dataset = load_dataset("Helsinki-NLP/opus-100", "en-sv", split="train[:5%]")

# with open(pretraining_, 'a') as f:
#     f.write(json.dumps({'metadata': {'dataset':"Helsinki-NLP/opus-100", "subset":"en-sv", 'split':'train[:5%]'},
#         'num_rows':dataset.num_rows}) + '\n')

# data = []

# for text in dataset:
#     data.append({'text': text['translation']['en'], 'metadata': {'dataset':"Helsinki-NLP/opus-100", "subset":"en-sv", 'split':'train[:5%]'}})

# with open(pretraining, 'a') as f:
#     for d in data:
#         f.write(json.dumps(d) + '\n')



from ..LLM_pieces import (
    RoPE,
    SMoE,
    SwiGLU,
    BertAttention,
    SlidingWindowAttention,
    create_static_block_mask,
    create_dynamic_block_mask,
    sliding_window_causal,
    DynamicTanh,
    Dynamic_erf
)
try:
    from cirilla.LLM_pieces import(
        MegablockMoE,
        MegablockdMoE,
    )
except ImportError:
    pass
from attn_gym.mods import generate_tanh_softcap
from dataclasses import dataclass
from .modules import select_torch_device
import torch.nn as nn
from typing import Optional
import warnings
import torch
import torch.nn.functional as F
from torchao.float8 import convert_to_float8_training, Float8LinearConfig
from torchao.sparsity.training import (
    SemiSparseLinear,
    swap_linear_with_semi_sparse_linear,
)

@dataclass
class EncoderArgs:
    """general"""
    dim:int = 256
    d_ff:int = 256
    n_layers:int = 2
    output_moe_weights:bool = False
    
    """attention"""
    context_window:int = 512 # max seq len
    n_heads:int = 2
    n_kv_heads:int = 2
    soft_cap:Optional[int] = 20

    """MoE"""
    num_experts:int = 2
    k:int = 1
    moe_type:str = "pytorch" # or "pytorch" or "megablocks-moe" or "megablocks-dmoe"
    moe_zloss_weight:float = 0.1
    capacity_factor: float = 1.0
    impl: str = "grouped"   # or "sparse" Sparse MLP is not supported with triton >=3.2.0
    
    """misc"""
    dtype_str:str = 'bfloat16'
    fp8_recipe:str="tensorwise" # tensorwise (fastest), rowwise, rowwise_with_gw_hp (most accurate)
    use_sparse:bool = False
    theta:float = 10_000.0
    device:str = select_torch_device()
    torch_compile:bool=True
    layer_norm:str = "RMSNorm" # or "Derf" or "DyT"

    @property
    def dtype(self):
        if self.dtype_str == "fp8":
            return torch.bfloat16 # for initialization, then convert to FP8
        return getattr(torch, self.dtype_str)

    def __post_init__(self):
        if not torch.cuda.is_available():
            warnings.warn("hf kernels only work on cuda")
        assert self.dim % self.n_heads == 0
        assert self.n_heads % self.n_kv_heads == 0
        if self.use_sparse:
            assert self.dtype_str != "fp8"
        if self.output_moe_weights:
            assert self.moe_type == "pytorch"

class Encoder(nn.Module):
    
    def __init__(self, args:EncoderArgs=None):
        super().__init__()

        if isinstance(args, dict):
            args = EncoderArgs(**args)

        if args is None:
            args = EncoderArgs()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):

        self.rope = RoPE(self.args.dim // self.args.n_heads, self.args.context_window, self.args.device, self.args.theta, self.args.device)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")

        self.attentions = [
            BertAttention(self.args, self.rope, generate_tanh_softcap(self.args.soft_cap, approx=False) if self.args.soft_cap is not None else None)
            for _ in range(self.args.n_layers)
            ]

        if self.args.dtype_str == "fp8":

            config = Float8LinearConfig.from_recipe_name(self.args.fp8_recipe)

            def module_filter_fn(mod: torch.nn.Module, fqn: str):
                # don't convert the last module
                if fqn == "1":
                    return False
                # don't convert linear modules with weight dimensions not divisible by 16
                if isinstance(mod, torch.nn.Linear):
                    if mod.in_features % 16 != 0 or mod.out_features % 16 != 0:
                        return False
                return True

            self.attentions = [convert_to_float8_training(attention, config=config, module_filter_fn=module_filter_fn) for attention in self.attentions]

        if self.args.use_sparse:

            def get_sparse_config(model, sparse_cls=SemiSparseLinear):
                config = {}
                for name, m in model.named_modules():
                    if isinstance(m, torch.nn.Linear):
                        out, inp = m.out_features, m.in_features
                        if out % 128 == 0 and inp % 128 == 0:
                            config[name] = sparse_cls
                return config
            
            for attention in self.attentions:
                swap_linear_with_semi_sparse_linear(attention, get_sparse_config(attention))
        
        if self.args.torch_compile:
            self.attentions = nn.ModuleList([
                torch.compile(attention.to(dtype=self.args.dtype), mode='max-autotune') for attention in self.attentions
                ])
        else:
            self.attentions = nn.ModuleList(self.attentions)
        
        if self.args.moe_type == 'pytorch':
            self.smoes = [
                SMoE(self.args, [SwiGLU(self.args) for _ in range(self.args.num_experts)])
                for _ in range(self.args.n_layers)
            ]

            if self.args.dtype_str == 'fp8':
                self.smoes = [convert_to_float8_training(smoe, config=config, module_filter_fn=module_filter_fn) for smoe in self.smoes]

            if self.args.use_sparse:
                for smoe in self.smoes:
                    swap_linear_with_semi_sparse_linear(smoe, get_sparse_config(smoe)) 

            if self.args.torch_compile:
                self.smoes = nn.ModuleList([
                    torch.compile(smoe.to(dtype=self.args.dtype), mode='max-autotune') for smoe in self.smoes
                ])
            else:
                self.smoes = nn.ModuleList(self.smoes)

        elif self.args.moe_type == 'megablocks-moe':
            self.smoes = nn.ModuleList([
                MegablockMoE(self.args)
                for _ in range(self.args.n_layers)
            ])

        elif self.args.moe_type == 'megablocks-dmoe':
            self.smoes = nn.ModuleList([
                MegablockdMoE(self.args)
                for _ in range(self.args.n_layers)
            ])
        
        else:
            print(self.args.moe_type)
            raise ValueError(f"allowed moe types: 'pytorch',  'megablocks-moe', 'megablocks-dmoe' ; got: {self.args.moe_type}")

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(dtype=self.args.dtype)
        
    def pred(self, x) -> torch.Tensor:
        
        if self.args.output_moe_weights:
            moe_weights = []

            for attention, moe in zip(self.attentions, self.smoes):

                x = x + attention(x)
                moe_out, moe_w = moe(x)
                moe_weights.append(moe_w)
                x = x + moe_out

            return x, moe_weights

        else:
            for attention, moe in zip(self.attentions, self.smoes):
                x = x + attention(x)
                x = x + moe(x)[0]
        
            return x
            
    def forward(self, x) -> torch.Tensor:
        return self.pred(x)


@dataclass
class DecoderArgs:
    """general"""
    dim:int = 1024
    d_ff:int = 2048
    n_layers:int = 16
    output_moe_weights:bool = False

    """attention"""
    context_window:int = 2048 # max seq len
    window_size:int = 1024
    n_heads:int = 8
    n_kv_heads:int = 4
    static_mask:bool = True
    soft_cap:Optional[int] = 20

    """MoE"""
    num_experts:int = 8
    k:int = 4
    moe_type:str = "pytorch" # or "pytorch" or "megablocks-moe" or "megablocks-dmoe"
    capacity_factor: float = 1.0
    moe_zloss_weight:float = 0.1
    impl: str = "grouped"   # or "sparse" Sparse MLP is not supported with triton >=3.2.0
    
    """misc"""
    dtype_str:str = 'bfloat16'
    fp8_recipe:str="tensorwise" # tensorwise (fastest), rowwise, rowwise_with_gw_hp (most accurate)
    use_sparse:bool = False
    theta:float = 10_000.0
    device = select_torch_device()
    torch_compile:bool=True
    layer_norm:str = "RMSNorm" # or "Derf" or "DyT"

    @property
    def dtype(self):
        if self.dtype_str == "fp8":
            return torch.bfloat16 # for initialization, then convert to FP8
        return getattr(torch, self.dtype_str)

    def __post_init__(self):
        if not torch.cuda.is_available():
            warnings.warn("hf kernels only work on cuda")
        assert self.dim % self.n_heads == 0
        assert self.n_heads % self.n_kv_heads == 0
        if self.use_sparse:
            assert self.dtype_str != "fp8"
        if self.output_moe_weights:
            assert self.moe_type == "pytorch"

class Decoder(nn.Module):

    def __init__(self, args:DecoderArgs=None):
        super().__init__()

        if isinstance(args, dict):
            args = DecoderArgs(**args)

        if args is None:
            args = DecoderArgs()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):

        self.rope = RoPE(self.args.dim // self.args.n_heads, self.args.context_window, self.args.device, self.args.theta, self.args.device)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
    
        if self.args.static_mask:
            self.mask = create_static_block_mask(sliding_window_causal,self.args.context_window,
                                            self.args.context_window, self.args.device, self.args.window_size)

            self.attentions = [
                SlidingWindowAttention(self.args, self.rope, self.mask, generate_tanh_softcap(self.args.soft_cap, approx=False) if self.args.soft_cap is not None else None)
                    for _ in range(self.args.n_layers)
            ]

        else:
            self.attentions = [
                SlidingWindowAttention(self.args, self.rope,
                create_dynamic_block_mask,
                generate_tanh_softcap(self.args.soft_cap, approx=False) if self.args.soft_cap is not None else None)
                    for _ in range(self.args.n_layers)
            ]

        if self.args.dtype_str == 'fp8':

            config = Float8LinearConfig.from_recipe_name(self.args.fp8_recipe)

            def module_filter_fn(mod: torch.nn.Module, fqn: str):
                # don't convert the last module
                if fqn == "1":
                    return False
                # don't convert linear modules with weight dimensions not divisible by 16
                if isinstance(mod, torch.nn.Linear):
                    if mod.in_features % 16 != 0 or mod.out_features % 16 != 0:
                        return False
                return True
            
            self.attentions = [convert_to_float8_training(attention, config=config, module_filter_fn=module_filter_fn) for attention in self.attentions]

        if self.args.use_sparse:

            def get_sparse_config(model, sparse_cls=SemiSparseLinear):
                config = {}
                for name, m in model.named_modules():
                    if isinstance(m, torch.nn.Linear):
                        out, inp = m.out_features, m.in_features
                        if out % 128 == 0 and inp % 128 == 0:
                            config[name] = sparse_cls
                return config
            
            for attention in self.attentions:
                swap_linear_with_semi_sparse_linear(attention, get_sparse_config(attention))

        if self.args.torch_compile:
            self.attentions = nn.ModuleList([
                torch.compile(attention.to(dtype=self.args.dtype), mode='max-autotune') for attention in self.attentions
                ])
        else:
            self.attentions = nn.ModuleList(self.attentions)
        
        if self.args.moe_type == 'pytorch':
            self.smoes = [
                SMoE(self.args, [SwiGLU(self.args) for _ in range(self.args.num_experts)])
                for _ in range(self.args.n_layers)
            ]

            if self.args.dtype_str == 'fp8':
                self.smoes = [convert_to_float8_training(smoe, config=config, module_filter_fn=module_filter_fn) for smoe in self.smoes]

            if self.args.use_sparse:
                for smoe in self.smoes:
                    swap_linear_with_semi_sparse_linear(smoe, get_sparse_config(smoe))        

            if self.args.torch_compile:
                self.smoes = nn.ModuleList([
                    torch.compile(smoe.to(dtype=self.args.dtype), mode='max-autotune') for smoe in self.smoes
                ])
            else:
                self.smoes = nn.ModuleList(self.smoes)

        elif self.args.moe_type == 'megablocks-moe':
            self.smoes = nn.ModuleList([
                MegablockMoE(self.args)
                for _ in range(self.args.n_layers)
            ])

        elif self.args.moe_type == 'megablocks-dmoe':
            self.smoes = nn.ModuleList([
                MegablockdMoE(self.args)
                for _ in range(self.args.n_layers)
            ])
        
        else:
            print(self.args.moe_type)
            raise ValueError(f"allowed moe types: 'pytorch',  'megablocks-moe', 'megablocks-dmoe' ; got: {self.args.moe_type}")

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(dtype=self.args.dtype)
        
    def pred(self, x) -> torch.Tensor:
        
        if self.args.output_moe_weights:
            moe_weights = []

            for attention, moe in zip(self.attentions, self.smoes):

                x = x + attention(x)
                moe_out, moe_w = moe(x)
                moe_weights.append(moe_w)
                x = x + moe_out

            return x, moe_weights

        else:
            for attention, moe in zip(self.attentions, self.smoes):
                x = x + attention(x)
                x = x + moe(x)[0]
        
            return x

    def forward(self, x) -> torch.Tensor:
        return self.pred(x)


@dataclass
class MixerArgs:
    """general"""
    dim: int = 256
    depth: int = 8
    context_window: int = 512
    expansion_factor: float = 4
    expansion_factor_token: float = 0.5
    dropout: float = 0.0
    dtype_str: str = 'bfloat16'
    device = select_torch_device()

    @property
    def dtype(self):
        return getattr(torch, self.dtype_str)

class MLPMixer1D(nn.Module):
    def __init__(self, args: Optional[MixerArgs] = None):
        super().__init__()

        if isinstance(args, dict):
            args = MixerArgs(**args)
        if args is None:
            args = MixerArgs()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):
        layers = []
        for _ in range(self.args.depth):
            # Token mixing
            layers.append(PreNormResidual(
                self.args.dim,
                FeedForward(self.args.context_window, int(self.args.expansion_factor * self.args.dim),
                                self.args.dropout, dense_type='conv1d')
            ))
            # Channel mixing
            layers.append(PreNormResidual(
                self.args.dim,
                FeedForward(self.args.dim, int(self.args.expansion_factor_token * self.args.dim),
                                self.args.dropout, dense_type='linear')
            ))

        self.norm = nn.LayerNorm(self.args.dim, bias = False)

        layers.append(self.norm)
        self.mixer = nn.Sequential(*layers)

        self.to(dtype=self.args.dtype, device=self.args.device)

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

    def pred(self, x: torch.Tensor) -> torch.Tensor:
        return self.mixer(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.pred(x)

class PreNormResidual(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim, bias = False)
        self.fn = fn

    def forward(self, x) -> torch.Tensor:
        return self.fn(self.norm(x)) + x

class FeedForward(nn.Module):
    def __init__(self, dim_in, dim_hidden, dropout=0., dense_type='linear'):
        super().__init__()
        if dense_type == 'conv1d':
            self.net = nn.Sequential(
                nn.Conv1d(dim_in, dim_hidden, kernel_size=1),
                nn.GELU(),
                nn.Dropout(dropout),
                nn.Conv1d(dim_hidden, dim_in, kernel_size=1),
                nn.Dropout(dropout)
            )
        elif dense_type == 'linear':
            self.net = nn.Sequential(
                nn.Linear(dim_in, dim_hidden),
                nn.GELU(),
                nn.Dropout(dropout),
                nn.Linear(dim_hidden, dim_in),
                nn.Dropout(dropout)
            )
        else:
            raise NotImplementedError

    def forward(self, x) -> torch.Tensor:
        x = self.net(x)
        return x

class PatchEmbed(nn.Module):
    def __init__(self, in_ch=3, embed_dim=768, patch_size=16):
        super().__init__()
        self.patch_size = patch_size
        self.proj = nn.Conv2d(in_ch, embed_dim, kernel_size=patch_size, stride=patch_size)
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x) -> torch.Tensor:
        # x ~ (B, C, H, W)
        x = self.proj(x)  # (B, D, H/ps, W/ps)
        x = x.flatten(2).transpose(1, 2)  # (B, H/ps * W/ps, D)
        x = self.norm(x)
        return x

class VisionEmbeddingModel(nn.Module):
    def __init__(self,
                    in_ch=3,
                    embed_dim=768,
                    patch_size=16,
                    H=16,
                    W=16
                    ):
        super().__init__()
        self.patch_embed = PatchEmbed(in_ch=in_ch, embed_dim=embed_dim, patch_size=patch_size)
        self.token_norm = nn.LayerNorm(embed_dim)
        self.pos_embed = self._make_pos_embed((H, W), embed_dim)

    def _make_pos_embed(self, grid_hw, embed_dim):
        H, W = grid_hw
        pe = nn.Parameter(torch.zeros(1, H * W, embed_dim))
        nn.init.trunc_normal_(pe, std=0.02)
        return pe

    def forward(self, x) -> torch.Tensor:
        # x ~ (B, C, H, W)
        tokens = self.patch_embed(x)  # (B, N, D)

        tokens = tokens + self.pos_embed
        tokens = self.token_norm(tokens)

        return tokens

class KeylessAttention(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.wi = nn.Linear(dim, 1, bias=False)
        self.wt = nn.Linear(dim, 1, bias=False)

    def forward(self, cls_image, cls_text) -> torch.Tensor:
        ei = self.wi(cls_image)
        et = self.wt(cls_text)
        lmb = F.sigmoid(ei - et)

        return lmb * cls_image + ((1 - lmb) * cls_text)

@dataclass
class EmbedArgs:
    vocab_size: int
    dim: int

class InputEmbeddings(nn.Module):
    def __init__(self, args: EmbedArgs):
        super().__init__()

        self.embeddings = nn.Embedding(args.vocab_size, args.dim)
    
    def forward(self, x) -> torch.Tensor:
        return self.embeddings(x)


import json
import os
import random

def gather_summaries(in_path, out_path):
    summary_paths = []
    for main_path, subfolders, files in os.walk(in_path):
        for file in files:
            if file.endswith(".txt"):
                summary_paths.append(os.path.join(main_path, file))
    
    out_data = []
    for path in summary_paths:
        with open(path, 'r') as f:
            summary = f.read()
        if not summary == '':
            out_data.append(
                    {'subject': path.split('/')[-1].split('.')[0], 'text': summary, 'data type': 'plain text', 'source':'fandom', 'model': path.split('/')[-2]}
            )

    with open(out_path, 'a') as f:
        for d in out_data:
            f.write(json.dumps(d) + '\n')

random_questions_to_summarize_texts = [
    "Tell me about {placeholder}.",
    "What is {placeholder}?",
    "Provide a brief overview of {placeholder}.",
    "Explain {placeholder} in simple terms.",
    "What are the key details about {placeholder}?",
    "Summarize all information about {placeholder}.",
    "Give me important facts about {placeholder}.",
    "What should I know about {placeholder}?",
    "Provide a short summary of {placeholder}.",
    "List the main points about {placeholder}.",
    "Give an introduction to {placeholder}.",
    "What is notable about {placeholder}?",
    "Describe {placeholder} in a few sentences.",
    "What makes {placeholder} significant?",
    "Provide an overview of {placeholder}.",
    "Give some key information about {placeholder}.",
    "What are the main aspects of {placeholder}?",
    "Provide background information on {placeholder}.",
    "Explain the significance of {placeholder}.",
    "What should someone know about {placeholder}?"
]
def summaries_to_instruct(in_path, out_path):
    summary_paths = []
    for main_path, subfolders, files in os.walk(in_path):
        for file in files:
            if file.endswith(".txt"):
                summary_paths.append(os.path.join(main_path, file))
    
    out_data = []
    for path in summary_paths:
        with open(path, 'r') as f:
            summary = f.read()
        if not summary == '':
            out_data.append(
                    {'subject': path.split('/')[-1].split('.')[0], 'text': [
                        {'role':'user', 'content': random.choice(random_questions_to_summarize_texts).format(placeholder=path.split('/')[-1].split('.')[0])},
                        {'role': 'assistant', 'content': summary}
                        ],
                        'data type': 'plain text', 'source':'fandom', 'model': path.split('/')[-2]}
            )

    with open(out_path, 'a') as f:
        for d in out_data:
            f.write(json.dumps(d) + '\n')



from .blocks import (
                    VisionEmbeddingModel,
                    KeylessAttention,
                    Encoder,
                    EncoderArgs,
                    InputEmbeddings
)
import torch.nn as nn
from huggingface_hub import PyTorchModelHubMixin
import torch
from ..LLM_pieces import SwiGLU, DynamicTanh, Dynamic_erf
from dataclasses import dataclass

@dataclass
class CMAArgs(EncoderArgs):
    in_channels:int = 3
    patch_size:int = 14
    H:int = 16
    W:int = 16
    n_tasks:int = 2
    n_classes:int = [2, 3]
    cls_text_index:int = 0
    cls_image_index:int = 10

class CMA(
        nn.Module,
        PyTorchModelHubMixin,
        pipeline_tag="text-generation",
        library_name="pytorch",
        license="mit"
    ):
    def __init__(self, args:CMAArgs=None):
        super().__init__()
        self.args = args
        self._prepare_model()

    def _prepare_model(self):
        
        self.vision_emb = VisionEmbeddingModel(self.args.in_channels,
                                                self.args.dim,
                                                self.args.patch_size,
                                                self.args.H,
                                                self.args.W
                                                )
        self.text_emb = InputEmbeddings(self.args)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
        self.encoder = Encoder(self.args)

        assert len(self.args.n_classes) == self.args.n_tasks
        self.outs = nn.ModuleList([nn.Sequential(
                                        KeylessAttention(self.args.dim),
                                        nn.Sequential(
                                            SwiGLU(self.args),
                                            nn.SiLU(),
                                            nn.Linear(self.args.dim, self.args.n_classes[i])
                                                )
                                            )
                                for i in range(self.args.n_tasks)])

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(self.args.device, dtype=self.args.dtype)

    def pred(self, texts, images, cls_image_token_index=None) -> list[torch.Tensor]:
        texts = self.text_emb(texts)
        images = self.vision_emb(images)
        if cls_image_token_index is not None:
            cls_i_em = self.text_emb(torch.tensor([cls_image_token_index]).to(self.args.device)).unsqueeze(0).expand(images.shape[0], -1, -1)
            x = torch.cat([texts, cls_i_em, images], dim=1)
        else:
            x = torch.cat([texts, images], dim=1)
        x = self.encoder(x)
        x = self.layer_norm(x)
        cls_text, cls_image = x[:, self.args.cls_text_index], x[:, self.args.cls_image_index]
        tasks = [out(cls_text, cls_image) for out in self.outs]
        return tasks


from cirilla.LLM_pieces import (
    RoPE,
    SMoE,
    SwiGLU,
    SlidingWindowAttention,
    create_static_block_mask,
    create_dynamic_block_mask,
    sliding_window_causal,
    DynamicTanh,
    Dynamic_erf
)
try:
    from cirilla.LLM_pieces import(
        MegablockMoE,
        MegablockdMoE,
    )
except ImportError:
    pass
from mamba_ssm import Mamba2
from cirilla.Cirilla_model.blocks import DecoderArgs
from attn_gym.mods import generate_tanh_softcap
from dataclasses import dataclass
import torch.nn as nn
import torch
from torchao.float8 import convert_to_float8_training, Float8LinearConfig
from torchao.sparsity.training import (
    SemiSparseLinear,
    swap_linear_with_semi_sparse_linear,
)

@dataclass
class HybridDecoderArgs(DecoderArgs):
    layer_pattern:str = 'MAMAM' # M - Mamba2 block, A - Attention block
    d_state:int = 64 # typically 64 or 128
    d_conv:int = 4

    def __post_init__(self):
        super().__post_init__()
        _layers=[]
        _number=''
        for layer in self.layer_pattern:
            if layer.isdigit():
                _number+=layer
            else:
                n_repeat = int(_number) if _number else 1
                _layers.append(layer*n_repeat)
                _number=''

        self.layer_pattern = "".join(_layers)

        if len(self.layer_pattern) != self.n_layers:
            raise ValueError(f"layer_pattern length {len(self.layer_pattern)} must match n_layers {self.n_layers}")
        
        assert self.d_ff % self.dim == 0, "d_ff must be multiple of dim"

class HybridDecoder(nn.Module):

    def __init__(self, args:HybridDecoderArgs=None):
        super().__init__()

        if isinstance(args, dict):
            args = HybridDecoderArgs(**args)

        if args is None:
            args = HybridDecoderArgs()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):

        self.rope = RoPE(self.args.dim // self.args.n_heads, self.args.context_window, self.args.device, self.args.theta, self.args.device)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
    
        if self.args.static_mask:
            self.mask = create_static_block_mask(sliding_window_causal,self.args.context_window,
                                            self.args.context_window, self.args.device, self.args.window_size)

        self.attentions = []

        for layer_type in self.args.layer_pattern:

            if layer_type == 'A':

                if self.args.static_mask:

                    self.attentions.append(
                        SlidingWindowAttention(self.args, self.rope, self.mask, generate_tanh_softcap(self.args.soft_cap, approx=False) if self.args.soft_cap is not None else None)
                    )

                else:
                    self.attentions.append(
                        SlidingWindowAttention(self.args, self.rope,
                            create_dynamic_block_mask,
                            generate_tanh_softcap(self.args.soft_cap, approx=False) if self.args.soft_cap is not None else None)
                    )
            
            elif layer_type == 'M':
                self.attentions.append(
                    Mamba2(
                        d_model = self.args.dim,
                        d_state = self.args.d_state,
                        d_conv = self.args.d_conv,
                        expand = self.args.d_ff // self.args.dim,
                        device = self.args.device,
                        dtype = self.args.dtype

                    )
                )
            else:
                raise ValueError(f"allowed layer types: 'A' for attention, 'M' for Mamba2 ; got: {layer_type}")

        if self.args.dtype_str == 'fp8':

            config = Float8LinearConfig.from_recipe_name(self.args.fp8_recipe)

            def module_filter_fn(mod: torch.nn.Module, fqn: str):
                # don't convert the last module
                if fqn == "1":
                    return False
                # don't convert linear modules with weight dimensions not divisible by 16
                if isinstance(mod, torch.nn.Linear):
                    if mod.in_features % 16 != 0 or mod.out_features % 16 != 0:
                        return False
                return True
            
            self.attentions = [convert_to_float8_training(attention, config=config, module_filter_fn=module_filter_fn) for attention in self.attentions]

        if self.args.use_sparse:

            def get_sparse_config(model, sparse_cls=SemiSparseLinear):
                config = {}
                for name, m in model.named_modules():
                    if isinstance(m, torch.nn.Linear):
                        out, inp = m.out_features, m.in_features
                        if out % 128 == 0 and inp % 128 == 0:
                            config[name] = sparse_cls
                return config
            
            for attention in self.attentions:
                swap_linear_with_semi_sparse_linear(attention, get_sparse_config(attention))

        if self.args.torch_compile:
            self.attentions = nn.ModuleList([
                torch.compile(attention.to(dtype=self.args.dtype), mode='max-autotune') for attention in self.attentions
                ])
        else:
            self.attentions = nn.ModuleList(self.attentions)
        
        if self.args.moe_type == 'pytorch':
            self.smoes = [
                SMoE(self.args, [SwiGLU(self.args) for _ in range(self.args.num_experts)])
                for _ in range(self.args.n_layers)
            ]

            if self.args.dtype_str == 'fp8':
                self.smoes = [convert_to_float8_training(smoe, config=config, module_filter_fn=module_filter_fn) for smoe in self.smoes]

            if self.args.use_sparse:
                for smoe in self.smoes:
                    swap_linear_with_semi_sparse_linear(smoe, get_sparse_config(smoe))        

            if self.args.torch_compile:
                self.smoes = nn.ModuleList([
                    torch.compile(smoe.to(dtype=self.args.dtype), mode='max-autotune') for smoe in self.smoes
                ])
            else:
                self.smoes = nn.ModuleList(self.smoes)

        elif self.args.moe_type == 'megablocks-moe':
            self.smoes = nn.ModuleList([
                MegablockMoE(self.args)
                for _ in range(self.args.n_layers)
            ])

        elif self.args.moe_type == 'megablocks-dmoe':
            self.smoes = nn.ModuleList([
                MegablockdMoE(self.args)
                for _ in range(self.args.n_layers)
            ])
        
        else:
            print(self.args.moe_type)
            raise ValueError(f"allowed moe types: 'pytorch',  'megablocks-moe', 'megablocks-dmoe' ; got: {self.args.moe_type}")

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(dtype=self.args.dtype)
        
    def pred(self, x) -> torch.Tensor:
        
        if self.args.output_moe_weights:
            moe_weights = []

            for attention, moe in zip(self.attentions, self.smoes):

                x = x + attention(x)
                moe_out, moe_w = moe(x)
                moe_weights.append(moe_w)
                x = x + moe_out

            return x, moe_weights

        else:
            for attention, moe in zip(self.attentions, self.smoes):
                x = x + attention(x)
                x = x + moe(x)[0]
        
            return x

    def forward(self, x) -> torch.Tensor:
        return self.pred(x)


from .Maml_trainer import MAMLBinaryAdapterTrainer, ReptileTrainer, MagMaxMAMLTrainer, Task, MamlPretrainingDataset
from .Protonet_trainer import ProtonetDataset, protonet_training_step, protonet_inference_step
from .Setfit_trainer import SetfitDataset, setfit_training_step, setfit_inference_step

__all__ = [
    "MAMLBinaryAdapterTrainer",
    "ReptileTrainer",
    "MagMaxMAMLTrainer",
    "Task",
    "MamlPretrainingDataset",
    "ProtonetDataset",
    "protonet_training_step",
    "protonet_inference_step",
    "SetfitDataset",
    "setfit_training_step",
    "setfit_inference_step"
]


from cirilla.LLM_pieces import DynamicTanh, Dynamic_erf
from dataclasses import dataclass
import torch.nn as nn
from .modules import CirillaBaseModel
from .blocks import Decoder, DecoderArgs, InputEmbeddings
import torch
from math import ceil

@dataclass
class Args(DecoderArgs):
    vocab_size:int = 60_000
    tie_params:bool = False
    out_bias:bool = False

class Cirilla(
            nn.Module,
            CirillaBaseModel,
            pipeline_tag="text-generation",
            library_name="pytorch",
            license="mit"
    ):
    def __init__(self, args:Args=None):
        super().__init__()

        if isinstance(args, dict):
            args = Args(**args)

        if args is None:
            args = Args()

        self.args = args
        self._prepare_model()

    def _prepare_model(self):

        self.emb = InputEmbeddings(self.args)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
        self.decoder = Decoder(self.args)

        self.output = nn.Linear(self.args.dim, self.args.vocab_size, bias=self.args.out_bias)
        if self.args.tie_params:
            self.output.weight = self.emb.embeddings.weight

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(self.args.device, dtype=self.args.dtype)
        
    def pred(self, x) -> torch.Tensor:
        
        x = self.emb(x)

        if self.args.output_moe_weights:
            x, moe_weights = self.decoder(x)

            x = self.layer_norm(x)
            x = self.output(x)

            return x, moe_weights
        
        else:
            x = self.decoder(x)

            x = self.layer_norm(x)
            x = self.output(x)
        
            return x

    def forward(self, x) -> torch.Tensor:
        return self.pred(x)
    
    @torch.no_grad()
    def infer_with_cache(self, x, cur_pos:int, max_batch:int=1, chunked_prefill:bool=False, non_finished_ids:torch.Tensor=None) -> torch.Tensor:
        
        x = self.emb(x)

        if self.args.output_moe_weights:

            for attention, moe in zip(self.decoder.attentions, self.decoder.smoes):

                x = x + attention.forward_with_cache(x, cur_pos, max_batch, chunked_prefill, non_finished_ids)
                moe_out, moe_weights = moe(x)
                x = x + moe_out

            x = self.layer_norm(x)
            x = self.output(x)

            return x
        
        else:

            for attention, moe in zip(self.decoder.attentions, self.decoder.smoes):
                x = x + attention.forward_with_cache(x, cur_pos, max_batch, chunked_prefill, non_finished_ids)
                x = x + moe(x)[0]

            x = self.layer_norm(x)
            x = self.output(x)
        
            return x

    @torch.no_grad()
    def infer(self, x) -> torch.Tensor:
        if self.args.output_moe_weights:
            logits, moe_weights = self.pred(x)
            return logits
        else:
            logits = self.pred(x)
            return logits
    
    def _greedy_next_token(self, x) -> torch.Tensor:
        logits = self.infer(x)
        probs = torch.nn.functional.softmax(logits[:, -1, :], dim=-1)
        next_token = torch.argmax(probs, dim=-1).unsqueeze(-1)
        return next_token
    
    def generate_naive(self, x:torch.Tensor,
                        max_new_tokens:int=1024,
                        top_k:int=None,
                        top_p:float=None,
                        n_beams:int=None,
                        temperature:float=1.0,
                        termination_tokens:list[int]=None
                        ) -> torch.Tensor:

        if top_k is None and top_p is None and n_beams is None: # pure greedy
            for _ in range(max_new_tokens):
                next_token = self._greedy_next_token(x)
                if termination_tokens is not None and next_token.item() in termination_tokens:
                    x = torch.cat((x, next_token), dim=1) # include termination token
                    break
                x = torch.cat((x, next_token), dim=1)
            return x
        
        else:

            with torch.no_grad():

                if n_beams is None:
                    n_beams = 1

                _beams = [[x, 0, False] for _ in range(n_beams)]

                for _ in range(max_new_tokens):
                    
                    n_remaining_top_p = None

                    if all([beam[2] for beam in _beams]): # all beams have reached termination
                        break
                    _new_beams = []

                    for beam in _beams:

                        if beam[2]: # termination already reached
                            _new_beams.append(beam)
                            continue

                        logits = self.infer(beam[0])
                        logits = logits[:, -1, :] / temperature

                        if top_k is not None:
                            values, indices = torch.topk(logits, top_k)
                            log_probs = torch.full_like(logits, float('-inf'))
                            log_probs = log_probs.scatter_(1, indices, torch.nn.functional.log_softmax(values, dim=-1))

                        elif top_p is not None:
                            sorted_logits, sorted_indices = torch.sort(logits, descending=True)
                            cumulative_probs = torch.cumsum(torch.nn.functional.softmax(sorted_logits, dim=-1), dim=-1)

                            sorted_indices_to_remove = cumulative_probs > top_p
                            sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
                            sorted_indices_to_remove[:, 0] = 0

                            indices_to_remove = sorted_indices[sorted_indices_to_remove]
                            n_remaining_top_p = logits.size(-1) - indices_to_remove.size(0)

                            log_probs = torch.nn.functional.log_softmax(logits, dim=-1)
                            log_probs[0, indices_to_remove] = float('-inf')

                        else: # greedy
                            values, indices = torch.topk(logits, n_beams)
                            log_probs = torch.full_like(logits, float('-inf'))
                            log_probs = log_probs.scatter_(1, indices, torch.nn.functional.log_softmax(values, dim=-1))

                        n_samples = min(n_beams,
                                        top_k if top_k is not None else float('inf'),
                                        n_remaining_top_p if n_remaining_top_p is not None else float('inf')
                                        )

                        next_tokens = torch.multinomial(log_probs.exp(), num_samples=n_samples, replacement=n_samples < n_beams) #batch_size x n_beams
                        next_tokens_probs = log_probs.gather(1, next_tokens)

                        for i in range(next_tokens.size(1)):

                            token = next_tokens[0, i].unsqueeze(0).unsqueeze(0)
                            token_prob = next_tokens_probs[0, i]

                            _new_beams.append([torch.cat([beam[0], token], dim=1),
                                                beam[1] + token_prob.item(),
                                                beam[2] or (termination_tokens is not None and token.item() in termination_tokens)
                                                ])

                    _beams = _new_beams
                
                    _beams = sorted(_beams, key=lambda x: x[1], reverse=True)[:n_beams]

                return _beams[0][0]

    def generate_kv_cache(self,
                            prompt_tokens_list: list[list[int]],
                            max_new_tokens: int = 1024,
                            top_k: int = None,
                            top_p: float = None,
                            temperature: float = 1.0,
                            termination_tokens: list[int] = None,
                            pad_token_id: int = 1
                            ) -> torch.Tensor:
            
            batch_size = len(prompt_tokens_list)
            
            prompt_lens = torch.tensor([len(t) for t in prompt_tokens_list], device=self.args.device)
            max_prompt_len = prompt_lens.max().item()
            min_prompt_len = prompt_lens.min().item()
            n_chunked_prefill_steps = ceil(min_prompt_len / self.args.window_size)
            
            total_len = min(self.args.context_window, max_prompt_len + max_new_tokens)
            
            tokens = torch.full((batch_size, total_len), pad_token_id, dtype=torch.long, device=self.args.device)
            
            for k, t in enumerate(prompt_tokens_list):
                tokens[k, :len(t)] = torch.tensor(t, dtype=torch.long, device=self.args.device)

            non_finished_ids = torch.arange(batch_size, device=self.args.device)
            
            with torch.inference_mode():
                
                cur_pos = 0

                for _ in range(n_chunked_prefill_steps):

                    chunk = tokens[:, cur_pos:min(cur_pos + self.args.window_size, min_prompt_len)]
                    logits = self.infer_with_cache(chunk, cur_pos=cur_pos, max_batch=batch_size, chunked_prefill=True)
                    next_token_logits = logits[:, -1, :]
                    cur_pos += chunk.shape[1]

                while cur_pos < total_len: # single token generation loop
                    
                    next_token_logits = next_token_logits / temperature

                    if top_k is not None:
                        v, i = torch.topk(next_token_logits, top_k)
                        probs = torch.full_like(next_token_logits, 0)
                        probs.scatter_(1, i, torch.nn.functional.softmax(v, dim=-1))

                    elif top_p is not None:
                        sorted_logits, sorted_indices = torch.sort(next_token_logits, descending=True)
                        cumulative_probs = torch.cumsum(torch.nn.functional.softmax(sorted_logits, dim=-1), dim=-1)

                        sorted_indices_to_remove = cumulative_probs > top_p
                        sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
                        sorted_indices_to_remove[:, 0] = 0
                        
                        mask = torch.zeros_like(next_token_logits, dtype=torch.bool).scatter_(1, sorted_indices, sorted_indices_to_remove)
                        next_token_logits[mask] = float('-inf')
                        probs = torch.nn.functional.softmax(next_token_logits, dim=-1)

                    else: # Greedy
                        max_arg = torch.argmax(next_token_logits, dim=-1) # (b,)
                        probs = torch.zeros_like(next_token_logits)
                        probs[range(probs.size(0)), max_arg] = 1.0

                    next_token_sample = torch.multinomial(probs, num_samples=1).squeeze(1) # (b,1) -> (b,)

                    is_prompt_phase = cur_pos < prompt_lens[non_finished_ids] # (b,)
                    ground_truth = tokens[non_finished_ids, cur_pos] # (b,)

                    next_token = torch.where(is_prompt_phase, ground_truth, next_token_sample) # (b,)

                    tokens[non_finished_ids, cur_pos] = next_token

                    if termination_tokens is not None:
                        active_generation_mask = ~is_prompt_phase
                        has_terminated = torch.isin(next_token, torch.tensor(termination_tokens, device=next_token.device, dtype=next_token.dtype)) & active_generation_mask
                        non_finished_ids = non_finished_ids[~has_terminated]
                    
                    if non_finished_ids.size(0) == 0:
                        break
                    
                    if termination_tokens is not None and has_terminated.any():
                        input_token = next_token[~has_terminated].unsqueeze(1) # add seq dim (b, 1)
                    else:
                        input_token = next_token.unsqueeze(1)

                    logits = self.infer_with_cache(input_token, cur_pos=cur_pos, non_finished_ids=non_finished_ids)
                    next_token_logits = logits[:, -1, :]
                    
                    cur_pos += 1

            return tokens[:, :cur_pos+1]
    
    def clear_cache(self) -> None:
        for att in self.decoder.attentions:
            att._clear_cache()



from cirilla.LLM_pieces import get_activation
import pytest
from requests import HTTPError

@pytest.mark.parametrize("activation_path", [
    "motif-technologies/optimizer"
])
def test_get_activation_succeeds(activation_path):
    result = get_activation(activation_path)
    assert result is not None

@pytest.mark.parametrize("activation_path", [
    "frog/frog-frogger-dogger",
])
def test_get_activation_fails(activation_path):
    with pytest.raises(HTTPError):
        get_activation(activation_path)



from torch.utils.data import Dataset
from ..Cirilla_model.dataloader import GenericDataset
import json
import torch
import random

class SetfitDataset(Dataset, GenericDataset):
    def __init__(self, tokenizer=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.tokenizer = tokenizer

        self._get_data()

    def __len__(self):
        return int((len(self.texts) * (len(self.texts)-1))/2)

    def _get_data(self):

        self.texts = []
        self.labels = []

        for p in self.path:
            with open(p, 'r') as f:

                for line in f:
                    line = json.loads(line)
                    assert line['data type'] == 'bert', 'Setfit supports only BERT data'

                    text = line['text']

                    if self.prefix_tokens is not None:
                        text = "".join(self.prefix_tokens) + text

                    if self.suffix_tokens is not None:
                        text += "".join(self.suffix_tokens)

                    self.texts.append(text)
                    self.labels.append(line['label'])

        slabels = set(self.labels)
        assert sorted(list(slabels)) == list(range(len(slabels))), "Labels must be consecutive integers starting from 0"
        self.labels = torch.tensor(self.labels, dtype=torch.int64)
        self.n_classes = len(slabels)

        for s in slabels:
            indices = torch.where(self.labels == s)[0]
            if len(indices) < 2:
                print(f"Label {s} has only {len(indices)} samples. Needs at least 2")
                continue

            setattr(self, f'indices_{int(s)}', indices.tolist())

    def __getitem__(self, idx):
                
        anchor_idx = torch.randint(0, self.n_classes, (1,))
        negative_idx = random.choice([i for i in range(self.n_classes) if i != anchor_idx])
        
        anchor_data = getattr(self, f'indices_{anchor_idx.item()}').copy()
        negative_data = getattr(self, f'indices_{negative_idx}').copy()

        _anchor_data = random.choice(anchor_data)
        anchor_data.remove(_anchor_data)
        positive_data = random.choice(anchor_data)
        n = random.choice(negative_data)

        anchor_tokens = self.tokenizer(self.texts[_anchor_data], return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
        positive_tokens = self.tokenizer(self.texts[positive_data], return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)
        negative_tokens = self.tokenizer(self.texts[n], return_tensors='pt', padding='max_length', truncation=True, max_length=self.max_len)

        anchor_encoded_id = anchor_tokens['input_ids'].to(self.device).squeeze(0)
        anchor_encoded_mask = anchor_tokens['attention_mask'].to(self.device).squeeze(0)

        positive_encoded_id = positive_tokens['input_ids'].to(self.device).squeeze(0)
        positive_encoded_mask = positive_tokens['attention_mask'].to(self.device).squeeze(0)

        negative_encoded_id =  negative_tokens['input_ids'].to(self.device).squeeze(0)
        negative_encoded_mask = negative_tokens['attention_mask'].to(self.device).squeeze(0)


        return (
            #'anchor': 
            (anchor_encoded_id, anchor_encoded_mask),
            #'positive':
            (positive_encoded_id, positive_encoded_mask),
            #'negative':
            (negative_encoded_id, negative_encoded_mask)
                )
    
def setfit_training_step(self, data):
    torch.compiler.cudagraph_mark_step_begin()

    anchor = data[0]
    positive = data[1]
    negative = data[2]

    anchor_out = self.model(*anchor)
    positive_out = self.model(*positive)
    negative_out = self.model(*negative)

    loss = self.criterion(anchor_out, positive_out, negative_out) # criterion has to be triplet loss
    loss_item = loss.item()
    loss.backward()
    return loss_item

@torch.inference_mode()
def setfit_inference_step(self, data):
    anchor = data[0]
    positive = data[1]
    negative = data[2]

    anchor_out = self.model(*anchor)
    positive_out = self.model(*positive)
    negative_out = self.model(*negative)

    loss = self.criterion(anchor_out, positive_out, negative_out) # criterion has to be triplet loss
    return loss.item()


from cirilla.Cirilla_model import Cirilla
from cirilla.LLM_pieces import DynamicTanh, Dynamic_erf
from .mamba_blocks import HybridDecoder, HybridDecoderArgs
from dataclasses import dataclass
import torch.nn as nn
from .blocks import InputEmbeddings


@dataclass
class HybridArgs(HybridDecoderArgs):
    vocab_size:int = 60_000
    tie_params:bool = False
    out_bias:bool = False

class HybridCirilla(Cirilla):

    def __init__(self, args:HybridArgs=None):
        super().__init__(args)

    def _prepare_model(self):

        self.emb = InputEmbeddings(self.args)
        if self.args.layer_norm == "RMSNorm":
            self.layer_norm = nn.RMSNorm(self.args.dim)
        elif self.args.layer_norm == "Derf":
            self.layer_norm = Dynamic_erf(self.args.dim)
        elif self.args.layer_norm == "DyT":
            self.layer_norm = DynamicTanh(self.args.dim)
        else:
            raise ValueError(f"allowed layer norms: 'RMSNorm', 'Derf', 'DyT' ; got: {self.args.layer_norm}")
        self.decoder = HybridDecoder(self.args)

        self.output = nn.Linear(self.args.dim, self.args.vocab_size, bias=self.args.out_bias)
        if self.args.tie_params:
            self.output.weight = self.emb.embeddings.weight

        self.n_params = sum(p.numel() for p in self.parameters() if p.requires_grad)

        self.to(self.args.device, dtype=self.args.dtype)